# DepthAI Gate
# Initial implementation in Python
# To be either ported or native modules added for interraction with USB

# Determines get script directory and cd into it
import os
SCRIPT_DIR = os.path.dirname(__file__) or '.'
os.chdir(SCRIPT_DIR)

from flask import Flask, jsonify, request, render_template, make_response, Blueprint
from flask_restful import Api, Resource, reqparse, fields, marshal, inputs
import werkzeug
from depthai_gate.exceptions import PublicError
from depthai_gate.session_manager import SessionManager, Session
from depthai_gate.fwp_manager import FwpManager
from depthai_gate.discovery import DeviceDiscovery, State, Platform
from depthai_gate.update_manager import UpdateManager
from depthai_gate import utilities
import logging
import subprocess
import os
import json
from flask import Response
from pathlib import Path

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = '0.0.0+local'

del PackageNotFoundError
del version

DEPHTAI_GATE_CONFIG_ENV_NAME = 'DEPTHAI_GATE_CONFIG'
DEFAULT_CONFIG_PATH = '/etc/depthai-gate.json'

# TODO(themarpe) - move these options to a configuration file
# Preferably /etc with the overlay
DEPTHAI_GATE_HTTP_PORT = 11492
DEPTHAI_GATE_HTTPS_PORT = 11492
DEPTHAI_GATE_HOST = '0.0.0.0'
DEPTHAI_GATE_DISCOVERY_HOST = '0.0.0.0'

BASE_API_URL = '/api/v1'
UPDATE_FOLDER_LOCATION = '/data/tmp/depthai_gate/'

class CustomApi(Api):
    """
    Patch Flask-style custom error handling into the Flask-RESTful api class.
    """
    def __init__(self, *args, **kwargs):
        super(CustomApi, self).__init__(*args, **kwargs)
        self._errorhandlers = []

    def errorhandler(self, exception_type):
        """
        Defined handlers for exceptions. Example:
        @api.errorhandler(ServerError):
        def handle_server_error(error):
            response = flask.jsonify({'message': error.message})
            response.status_code = error.status_code
            return response
        """
        def wrapper(func):
            import functools
            self._errorhandlers.append((exception_type, func))
            # Sort error handlers to have sub exceptions first, so that those
            # take preference over base exceptions.

            self._errorhandlers = sorted(
                self._errorhandlers,
                key=functools.cmp_to_key(self._inheritance_comparator))
            logging.error(self._errorhandlers)
            return func
        return wrapper

    def handle_error(self, error, previous_errors=None):
        # Keep track of previous errors in the current chain of exception
        # handling in order to prevent infinite cycles that would occur if two
        # error handler raise the exception handled by the other.
        logging.error(f'Exception: {error}')
        previous_errors = previous_errors or []
        previous_errors.append(type(error))
        # Try to find the first custom handler for the occured exception.
        for exception_type, handler in self._errorhandlers:
            if not isinstance(error, exception_type):
                continue
            try:
                return handler(error)
            except Exception as new_error:
                if type(new_error) not in previous_errors:
                    return self.handle_error(new_error, previous_errors)
            break
        # If no matching handler was found or an infinite cycle is detected,
        # fall back to Flask-RESTful's error handling.
        return super(CustomApi, self).handle_error(error)

    @staticmethod
    def _inheritance_comparator(lhs, rhs):
        lhs_sub = issubclass(lhs, rhs)
        rhs_sub = issubclass(lhs, rhs)
        if lhs_sub and not rhs_sub:
            return -1
        if rhs_sub and not lhs_sub:
            return 1
        return 0

def main():
    # logging.basicConfig(encoding='utf-8', level=logging.DEBUG)
    logging.basicConfig(level=logging.DEBUG)
    logging.debug('DepthAI Gate - Debug Logging ON!')
    # First, detect the platform
    platform = Platform.KEEMBAY
    # TODO(Morato) - check if this is reliable
    if os.path.exists('/sys/devices/platform/keembay_camera'):
        platform = Platform.KEEMBAY
    elif os.path.exists('/sys/module/qcom_ipcc'):
        platform = Platform.RVC4
    else:
        logging.error('Could not detect platform, defaulting to KEEMBAY')
        platform = Platform.KEEMBAY

    # Query ENV & settings
    config_path = DEFAULT_CONFIG_PATH
    config_path_set = False
    if DEPHTAI_GATE_CONFIG_ENV_NAME in os.environ:
        config_path = os.environ[DEPHTAI_GATE_CONFIG_ENV_NAME]
        config_path_set = True

    config = {
        'httpPort': DEPTHAI_GATE_HTTP_PORT,
        'httpsPort': DEPTHAI_GATE_HTTPS_PORT,
        'host': DEPTHAI_GATE_HOST,
        'discoveryHost': DEPTHAI_GATE_DISCOVERY_HOST,
        'loggingLevel': 'WARN'
    }

    # Update config with given one
    try:
        with open(config_path, "rb") as f:
            config_json = json.load(f)
            for key in config.keys() & config_json.keys():
                config[key] = config_json[key]
    except Exception as ex:
        if config_path_set:
            logging.error(f'Warning, could not apply custom config, ex: {ex}')
        else:
            logging.info(f'No default config at {DEFAULT_CONFIG_PATH} found. Using defaults...')
    # Get the logging level from config
    logging_level = config['loggingLevel']
    try:
        logging.getLogger().setLevel(logging_level)
    except TypeError as ex:
        logging.getLogger().setLevel(logging.WARN)
        logging.error(f'Invalid logging level {logging_level} - exception {ex}, using WARN')
    logging.debug(f'Config: {config}')

    app = Flask(__name__)
    app.config['MAX_CONTENT_LENGTH'] = 1 * 1024 * 1024 * 1024 # 1GiB for OS image upload
    app.config['UPLOAD_FOLDER'] = '/data' # As more storage is available
    blueprint = Blueprint('api', __name__, url_prefix=BASE_API_URL)
    api = CustomApi(blueprint)
    app.register_blueprint(blueprint)

    # Query OS and Gate versions once
    version_os = subprocess.check_output(f'cat /etc/os-release | grep "VERSION_ID"', shell=True).decode('utf-8').strip().split('=')[1].strip('"')
    version_gate = __version__


    @api.errorhandler(PublicError)
    def handle_exception(err):
        """Return JSON instead of HTML for MyCustomError errors."""
        response = {
            "error": str(err),
        }
        return jsonify(response), err.code

    session_fields = {
        "id": fields.String,
        "name": fields.String,
        "fwp_checksum": fields.String,
        "fwp_version": fields.String,
        "library_version": fields.String,
        # "environment": fields.String,
        "connection_information": fields.String,
        "owner": fields.String,
        "protected": fields.Boolean,
        "create_datetime": fields.DateTime('iso8601'),
        "start_datetime": fields.DateTime('iso8601'),
        "stop_datetime": fields.DateTime('iso8601'),
        "destroy_datetime": fields.DateTime('iso8601'),
        "fwp_exists": fields.Boolean,
        "state": fields.String,
    }

    fwp_manager = FwpManager(platform)
    session_manager = SessionManager(fwp_manager)
    update_manager = UpdateManager()

    # /sessions
    class SessionListAPI(Resource):
        def __init__(self, manager, fwp_manager) -> None:
            super().__init__()
            self.manager = manager
            self.fwp_manager = fwp_manager
            self.parse_create_session = reqparse.RequestParser()
            self.parse_create_session.add_argument('name', type=str, required=False, location='json')
            self.parse_create_session.add_argument('fwp_checksum', type=str, required=False, location='json')
            self.parse_create_session.add_argument('fwp_version', type=str, required=False, location='json')
            self.parse_create_session.add_argument('library_version', type=str, required=True, help='No library version specified', location='json')
            self.parse_create_session.add_argument('environment', type=str, required=False, location='json')
            self.parse_create_session.add_argument('protected', type=bool, required=False, location='json')

        def get(self):
            return {'sessions': [marshal(session, session_fields) for session in self.manager.get_all()]}

        def post(self):
            args = self.parse_create_session.parse_args()
            session = self.manager.create(args['name'], args['fwp_checksum'], args['fwp_version'], args['library_version'], args['environment'], '', '', args['protected'])
            return marshal(session, session_fields)
    api.add_resource(SessionListAPI, '/sessions', resource_class_kwargs={'manager': session_manager, 'fwp_manager': fwp_manager})

    # /sessions/{id}
    class SessionAPI(Resource):
        def __init__(self, manager) -> None:
            super().__init__()
            self.manager = manager
        def get(self, id):
            logging.debug(f"Type of ID  is {type(id)}")
            return marshal(self.manager.get(id), session_fields)
        def delete(self, id):
            self.manager.delete(id)
            return {'result': True}
    api.add_resource(SessionAPI, '/sessions/<string:id>', resource_class_kwargs={'manager': session_manager})

    # /sessions/{id}/fwp
    class SessionFwpAPI(Resource):
        def __init__(self, manager, fwp_manager) -> None:
            super().__init__()
            self.manager = manager
            self.fwp_manager = fwp_manager
            self.upload_parser = reqparse.RequestParser()
            # self.upload_parser.add_argument('', type=bytes, required=True, location='data')
            self.upload_parser.add_argument('dataUrl', type=inputs.boolean, default=False)
        def post(self, id):
            parse = reqparse.RequestParser()
            parse.add_argument('file', type=werkzeug.datastructures.FileStorage, location='files')
            args = parse.parse_args()
            fwp_filestorage = args['file']
            fwp_binary = fwp_filestorage.read()
            logging.debug(f"Size of the FWP binary is {len(fwp_binary)}")
            checksum = self.fwp_manager.create(fwp_binary)
            self.manager.get(id).fwp_checksum = checksum

            # Check for hailo compatibility
            binary_path = self.fwp_manager.get_entry_point(checksum)
            if utilities.hailo_compatible(binary_path):
                return {'result': True}
            else:
                return {'result': False, 'message': 'Depthai is not compatible with Hailo - either depthai needs to be updated or OS downgraded'}, 422
    api.add_resource(SessionFwpAPI, '/sessions/<string:id>/fwp', resource_class_kwargs={'manager': session_manager, 'fwp_manager': fwp_manager})

    # /sessions/{id}/start
    class SessionStartAPI(Resource):
        def __init__(self, manager, fwp_manager) -> None:
            super().__init__()
            self.manager = manager
            self.fwp_manager = fwp_manager
        def post(self, id):
            self.manager.start(id)
            return {'result': True}
    api.add_resource(SessionStartAPI, '/sessions/<string:id>/start', resource_class_kwargs={'manager': session_manager, 'fwp_manager': fwp_manager})

    # /sessions/{id}/stop
    class SessionStopAPI(Resource):
        def __init__(self, manager, fwp_manager) -> None:
            super().__init__()
            self.manager = manager
            self.fwp_manager = fwp_manager
        def post(self, id):
            self.manager.stop(id)
            return {'result': True}
    api.add_resource(SessionStopAPI, '/sessions/<string:id>/stop', resource_class_kwargs={'manager': session_manager, 'fwp_manager': fwp_manager})

    class SessionDestroyAPI(Resource):
        def __init__(self, manager, fwp_manager) -> None:
            super().__init__()
            self.manager = manager
            self.fwp_manager = fwp_manager
        def post(self, id):
            self.manager.destroy(id)
            return {'result': True}
    api.add_resource(SessionDestroyAPI, '/sessions/<string:id>/destroy', resource_class_kwargs={'manager': session_manager, 'fwp_manager': fwp_manager})

    # /sessions/{id}/core_dump
    class SessionCoreDumpAPI(Resource):
        def __init__(self, manager) -> None:
            super().__init__()
            self.manager = manager

        def get(self, id: str):
            session: Session = self.manager.get(id)
            compressed_file_path : os.path = session.get_core_dump_all_compressed()
            if compressed_file_path is not None:
                headers = {"X-Filename": os.path.basename(compressed_file_path)}
                try:
                    with open(compressed_file_path, 'rb') as f:
                        data = f.read()
                        return Response(data, mimetype='application/octet-stream', headers=headers)
                except Exception as e:
                    logging.error(f"Error while reading core dump file: {e}")
                    return {'result': None}, 404
            else:
                return {'result': None}, 404
    api.add_resource(SessionCoreDumpAPI, '/sessions/<string:id>/core_dump', resource_class_kwargs={'manager': session_manager})

    # /status
    class StatusAPI(Resource):
        def get(self):
            return {'status': True}
    api.add_resource(StatusAPI, '/status')

    # /status
    class VersionAPI(Resource):
        def get(self):
            return {'version_gate': version_gate, 'version_os': version_os}
    api.add_resource(VersionAPI, '/version')

    # /update/os
    class UpdateOsAPI(Resource):
        def __init__(self, manager):
            self.manager = manager

        def post(self):
            """Update OS"""

            # Save the uploaded file
            f = request.files['file']
            file_location = '/data/tmp/depthai_gate/os_update.mender'
            folder_location = UPDATE_FOLDER_LOCATION
            os.makedirs(folder_location, exist_ok = True)
            f.save(file_location)
            logging.debug('Upload complete, OS update image received. Installing...')

            # Start update (if already running, raises an error)
            self.manager.update(file_location)

            # Return a progress script for better UX
            response = make_response(render_template('update_progress.html', api=BASE_API_URL))
            response.content_type = 'text/HTML'
            return response

        def get(self):
            """Retrieve OS Update progress"""
            return {'inprocess': self.manager.inprocess, 'progress': self.manager.progress, 'last_message': self.manager.last_message}
    api.add_resource(UpdateOsAPI, '/update/os', resource_class_kwargs={'manager': update_manager})


    ## Non API part
    # Index page
    @app.route('/')
    def index():
        # Return a progress script for better UX
        response = make_response(render_template('index.html', api=BASE_API_URL, version_os=version_os, version_gate=version_gate))
        response.content_type = 'text/HTML'
        return response

    # retrieve serial
    # /sys/firmware/devicetree/base/serial-number
    # MX: 18443010818C0E0900
    # KB: 2b528816838860c6
    # KB: 2b08a816884861c2
    serial = 'xlink-gate'
    try:
        if platform == Platform.KEEMBAY:
            with open('/sys/firmware/devicetree/base/serial-number') as f:
                serial = f.readlines()[0].split('\x00')[0]
        elif platform == Platform.RVC4:
            with open('/sys/bus/soc/devices/soc0/serial_number') as f:
                serial = f.readlines()[0].split('\n')[0]
        else:
            raise PublicError(f'Unsupported platform: {platform}')
    except Exception as ex:
        logging.error(f"Couldn't retrieve devices serial number. Using default '{serial}'")


    def get_state_cb():
        active_session = False
        sessions = session_manager.get_all()
        for session in sessions:
            if session.state not in ['STOPPED', 'CRASHED', 'DESTROYED']:
                active_session = True
        if active_session:
            return State.GATE_BOOTED
        # Check if device is in setup mode
        factoryFileExists = Path("/persist/factory/enabled").exists()
        setupFileExists = Path("/etc/agentconfd/.setup_complete").exists()
        if not factoryFileExists and not setupFileExists:
            return State.GATE_SETUP
        return State.GATE

    discovery = DeviceDiscovery(serial, portHttp=config['httpPort'], portHttps=config['httpsPort'], address=config['discoveryHost'], platform=platform)
    discovery.set_state_cb(get_state_cb)
    # app.run(port=DEPTHAI_GATE_HTTPS_PORT, ssl_context='adhoc')
    app.run(host=config['host'], port=config['httpPort'])

if __name__ == '__main__':
    main()
