import hashlib
from pathlib import Path
import tempfile
import subprocess
import os
import logging
from depthai_gate.exceptions import PublicError
from depthai_gate.discovery import Platform
from threading import RLock
import glob

class Fwp:
    checksum: str
    package: bytes
    entry: str
    extract_path: Path
    binary_name: str

    def __init__(self, fwp_binary, checksum, extract_path, platform: Platform) -> None:
        self.package = fwp_binary
        self.checksum = checksum
        self.extract_path = Path(extract_path)
        if platform == Platform.KEEMBAY:
            self.binary_name = 'depthai-device-kb'
        elif platform == Platform.RVC4:
            self.binary_name = 'depthai-device-rvc4'
        else:
            raise PublicError(f'Unsupported platform: {platform}')

    # TODO(themarpe)
    # def extract_memory(self):
    #     tarfile.op
    #     lzma.decompress(self.package)
    #     with lzma.open("file.tar.xz") as fd:
    #         with tarfile.open(fileobj=fd) as tar:
    #             content = tar.extractall('/path/to/extract/to

    def extract_file(self, package_path):
        # secure_temp_dir = tempfile.TemporaryDirectory(prefix="fwp_")
        os.makedirs(self.extract_path / Path(self.checksum))
        subprocess.check_call(['tar', 'xf', package_path], cwd=self.extract_path / self.checksum)

    def exists(self):
        return (self.extract_path / self.checksum / self.binary_name).exists()

    def get_entry_point(self):
        path = self.extract_path / self.checksum / self.binary_name
        if not self.exists():
            raise PublicError('Entry point does not exist, make sure FWP was extracted first')
        return path

    def has_vpu_fw(self):
        path = self.extract_path / self.checksum / 'resources' / 'depthai-vpu-fw-kb-*'
        path = glob.glob(str(path))
        return len(path) == 1

    def get_root_folder(self):
        return self.extract_path / self.checksum

    def get_vpu_fw(self):
        path = self.extract_path / self.checksum / 'resources' / 'depthai-vpu-fw-kb-*'
        path = glob.glob(str(path))
        if len(path) != 1:
            raise PublicError('VPU firmware does not exist, make sure FWP was extracted first')
        return path[0]

class FwpManager:
    fwp_map: dict
    extract_dir: Path
    package_dir: Path
    extract_tempdir: tempfile.TemporaryDirectory
    package_tempdir: tempfile.TemporaryDirectory
    mutex: RLock
    platform: Platform

    def __init__(self, platform: Platform) -> None:
        self.extract_tempdir = tempfile.TemporaryDirectory(prefix="gate_fw_")
        self.package_tempdir = tempfile.TemporaryDirectory(prefix="gate_package_")
        self.extract_dir = Path(self.extract_tempdir.name)
        self.package_dir = Path(self.package_tempdir.name)
        self.fwp_map = {}
        self.mutex = RLock()
        self.platform = platform
        logging.debug(f'[FwpManager] extract_dir: {self.extract_dir}, package_dir: {self.package_dir}')

    def create(self, fwp_binary):
        with self.mutex:
            fwp_checksum = hashlib.sha256(fwp_binary).hexdigest()
            if fwp_checksum in self.fwp_map:
                if(self.fwp_map[fwp_checksum].exists()):
                    return fwp_checksum
                else:
                    del self.fwp_map[fwp_checksum]
            os.makedirs(self.extract_dir, exist_ok=True)
            os.makedirs(self.package_dir, exist_ok=True)
            self.fwp_map[fwp_checksum] = Fwp(fwp_binary, fwp_checksum, self.extract_dir, platform=self.platform)
            package_path = self.package_dir.absolute() / f'{fwp_checksum}.tar.xz'
            logging.debug(f'[FwpManager] package_path {package_path}')
            # TODO(themarpe) - extract directly from memory instead
            with open(package_path, 'wb') as package:
                package.write(fwp_binary)
            self.fwp_map[fwp_checksum].extract_file(package_path)
            return fwp_checksum

    def get_entry_point(self, fwp_checksum):
        with self.mutex:
            if fwp_checksum not in self.fwp_map:
                raise PublicError(f'FWP with a given checksum: {fwp_checksum} does not exist')
            return self.fwp_map[fwp_checksum].get_entry_point()

    def get_root_folder(self, fwp_checksum):
        with self.mutex:
            if fwp_checksum not in self.fwp_map:
                raise PublicError(f'FWP with a given checksum: {fwp_checksum} does not exist')
            return self.fwp_map[fwp_checksum].get_root_folder()

    def has_vpu_fw(self, fwp_checksum):
        with self.mutex:
            if fwp_checksum not in self.fwp_map:
                raise PublicError(f'FWP with a given checksum: {fwp_checksum} does not exist')
            return self.fwp_map[fwp_checksum].has_vpu_fw()

    def get_vpu_fw(self, fwp_checksum):
        with self.mutex:
            if fwp_checksum not in self.fwp_map:
                raise PublicError(f'FWP with a given checksum: {fwp_checksum} does not exist')
            return self.fwp_map[fwp_checksum].get_vpu_fw()

    def get_all(self):
        with self.mutex:
            return self.fwp_map

    def exists(self, checksum):
        with self.mutex:
            return checksum in self.fwp_map



# class FwpManagerRes(Resource):

#     def __init__(self, manager) -> None:
#         super().__init__()
#         self.last_unique_id = 0
#         self.manager = manager

#     # REST
#     def get(self):
#         return json.dumps(self.manager.session_map)
#     def get(self, )

#     def post(self):
#         self.manager.last_unique_id = self.manager.last_unique_id + 1
#         return json.dumps(self.manager.last_unique_id)
#         # data = self.request.get_json()
#         # TODO(themarpe) - connection information, owner, protected
#         # self.create(data['name'], data['fwp_checksum'], data['fwp_version'], data['library_version'], data['environment'], '', '', false)
