from depthai_gate.fwp_manager import FwpManager
from depthai_gate.exceptions import PublicError
from depthai_gate.watchdog import Watchdog

from datetime import datetime
import subprocess
from flask_restful import Resource
import json
import logging
from threading import RLock
import os
import importlib.util
import sys
import resource
import threading
import collections
import tarfile
import time
from pathlib import Path
TIMEOUT_BEFORE_DESTORY_AFTER_CREATE = 15 # Seconds


class OutputCollector(threading.Thread):
    def __init__(self, process, file_path, max_lines=10000):
        threading.Thread.__init__(self)
        self.process = process
        self.output = collections.deque(maxlen=max_lines)
        # Check if the file can be created
        try:
            with open(file_path, 'w') as f:
                pass
        except Exception as e:
            raise PublicError(f"Cannot open a file for depth logs {file_path}: {e}")
        self.file_path = file_path
    def run(self):
        for line in iter(self.process.stdout.readline, b''):
            self.output.append(line.decode('utf-8'))
            logging.debug(line.decode('utf-8'))
        logging.debug(f"Process {self.process.pid} finished - dumping the logs to a file")
        self.dump_to_file()

    def dump_to_file(self):
        with open(self.file_path, 'w') as f:
            f.write("\n".join(self.output))


class Session:
    id: str
    name: str
    fwp_checksum: str
    fwp_version: str
    library_version: str
    environment: dict
    connection_information: str
    owner: str
    protected: bool
    proc: subprocess.Popen
    create_datetime: datetime
    start_datetime: datetime
    stop_datetime: datetime
    destroy_datetime: datetime
    fwp_exists: bool
    watchdog: Watchdog
    state: str # CREATED, RUNNING, STOPPED, STOPPING, CRASHED, DESTROYED
    output_collector: OutputCollector
    log_file_path: Path
    core_dump_path: Path


    def __init__(self, id, name, fwp_checksum, fwp_version, library_version, environment, connection_information, create_datetime, start_datetime, stop_datetime, destroy_datetime, owner, protected):
        self.id = id
        self.name = name
        self.fwp_checksum = fwp_checksum
        self.fwp_version = fwp_version
        self.library_version = library_version
        self.environment = environment
        self.connection_information = connection_information
        self.create_datetime = create_datetime
        self.start_datetime = start_datetime
        self.stop_datetime = stop_datetime
        self.destroy_datetime = destroy_datetime
        self.owner = owner
        self.protected = protected
        self.proc = None
        self.fwp_exists = False
        self.watchdog = Watchdog(id)
        self.output_collector = None
        self.state = 'CREATED'
        self.log_file_path = None
        self.core_dump_path = None
        self.tar_file_path = None
        self.status_thread = threading.Thread(target=self.status_thread_run, daemon=True)
        self.status_thread.start()

    def __del__(self):
        for path in [self.log_file_path, self.core_dump_path, self.tar_file_path]:
            if path:
                if os.path.exists(path):
                    try:
                        os.remove(path)
                    except Exception as e:
                        logging.error(f"Cannot delete file {path}: {e}")

    def status_thread_run(self):
        t_at_start = time.time()
        while self.state == 'CREATED':
            time.sleep(1)
            if time.time() - t_at_start > TIMEOUT_BEFORE_DESTORY_AFTER_CREATE:
                logging.info(f"Session {self.id} created but not started - destroying")
                self.state = 'DESTROYED'
                break
    def get_core_dump_path(self):
        if not self.core_dump_path:
            logging.info("Core dump path not set")
            return None
        if not os.path.exists(self.core_dump_path):
            logging.info(f"Core dump file does not exist {self.core_dump_path}")
            return None
        return self.core_dump_path
    def get_logs_path(self):
        if not self.log_file_path:
            logging.info("Log file path not set")
            return None
        if not os.path.exists(self.log_file_path):
            logging.info(f"Log file does not exist {self.log_file_path}")
            return None
        return self.log_file_path
    # Returns path to the tarred and compressed file
    def get_core_dump_all_compressed(self):
        file_paths = []
        log_path = self.get_logs_path()
        if log_path:
            file_paths.append(log_path)
        core_dump_path = self.get_core_dump_path()
        if core_dump_path:
            file_paths.append(core_dump_path)
        # Check if the kernel log path exits
        kernel_log_path = Path("/var/log/kern.log")
        if kernel_log_path.exists():
            logging.info(f"Kernel log file exists {kernel_log_path}")
            file_paths.append(kernel_log_path)
        else:
            logging.warn(f"Kernel log file does not exist {kernel_log_path}")

        dmesg_path = Path("/tmp/dmesg.log")
        try:
            with open(dmesg_path, 'w') as f:
                subprocess.check_call(['dmesg'], stdout=f)
            file_paths.append(dmesg_path)
        except Exception as e:
            logging.error(f"Cannot get dmesg logs {e}")

        logcat_path = Path("/tmp/logcat.log")
        try:
            with open(logcat_path, 'w') as f:
                subprocess.check_call(['logcat', '-d'], stdout=f)
            file_paths.append(logcat_path)
        except Exception as e:
            logging.warning(f"Cannot get logcat logs {e}")
        os_release_path = Path("/etc/os-release")
        os_release_txt = Path("/tmp/os-release.txt")
        if os_release_path.exists():
            with open(os_release_path, 'r') as f:
                with open(os_release_txt, 'w') as f2:
                    for line in f:
                        f2.write(line)
            file_paths.append(os_release_txt)
        else:
            logging.warning(f"Cannot find os-release file {os_release_path}")

        device_model_path = Path("/proc/device-tree/model")
        if device_model_path.exists():
            file_paths.append(device_model_path)
        # Tar and compress all files
        if len(file_paths) == 0:
            logging.info("No files to compress")
            return None
        logging.info(f"Compressing files {file_paths}")

        try:
            # Create a tar file
            tar_file_name = f"core_dump_{self.fwp_version}_{time.strftime('%Y%m%d_%H%M%S')}.tar.gz"
            base_path = os.path.dirname(file_paths[0])
            self.tar_file_path = os.path.join(base_path, tar_file_name)
            with tarfile.open(self.tar_file_path, "w:gz", compresslevel=1) as tar:
                for file_path in file_paths:
                    tar.add(file_path, arcname=os.path.basename(file_path))
            logging.info(f"Files compressed into {self.tar_file_path}")
        except Exception as e:
            logging.error(f"Cannot compress files {file_paths}: {e}")
            return None
        return self.tar_file_path


class SessionManager():
    last_unique_id: int
    fwp_manager: FwpManager
    mutex: RLock
    on_event_module = None
    # session_map: Dict[str, Session]

    def load_module(self, module):

        spec = importlib.util.spec_from_file_location("module.name", str(module))
        m = importlib.util.module_from_spec(spec)
        sys.modules["module.name"] = m
        spec.loader.exec_module(m)

        return m


    def __init__(self, fwp_manager) -> None:
        super().__init__()
        self.last_unique_id = 0
        self.session_map = dict()
        self.fwp_manager = fwp_manager
        self.mutex = RLock()

    def _get_unique_session_id(self):
        with self.mutex:
            self.last_unique_id = self.last_unique_id + 1
            return str(self.last_unique_id)

    def create(self, name, fwp_checksum, fwp_version, library_version, environment, connection_information, owner, protected):
        with self.mutex:
            # Delete all current session - multiple sessions not supported for now
            for session in self.session_map.values():
                self.destroy(session.id)
            self.session_map.clear()

            session = Session(self._get_unique_session_id(), name, fwp_checksum, fwp_version, library_version, environment, connection_information, datetime.now(), None, None, None, owner, protected)
            self.session_map[session.id] = session
            self.fwp_exists = self.fwp_manager.exists(session.fwp_checksum)
            return session

    def start(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            session = self.session_map[id]

            session.start_datetime = datetime.now()

            entry = self.fwp_manager.get_entry_point(session.fwp_checksum)
            root_folder = self.fwp_manager.get_root_folder(session.fwp_checksum)

            script_folder = root_folder / "scripts"
            custom_script = script_folder / "custom_scripts.py"

            try:
                self.on_event_module = self.load_module(str(custom_script))
            except Exception as e:
                logging.error("Exception thrown on loading the module for the custom script {e}")
                self.on_event_module = None
            custom_env = {}
            if self.on_event_module:
                try:
                    custom_env = self.on_event_module.on_start()
                except Exception as e:
                    logging.error("Exception thrown on calling the on_start function {e}")

            has_vpu_fw = self.fwp_manager.has_vpu_fw(session.fwp_checksum)
            if 'DEPTHAI_GATE_SKIP_VPU_FW_BOOT' in os.environ and os.environ['DEPTHAI_GATE_SKIP_VPU_FW_BOOT'] != '0':
                logging.warn('Skipping booting VPU FW due to DEPTHAI_GATE_SKIP_VPU_FW_BOOT being set')
                has_vpu_fw = False
            vpu_fw = None
            if has_vpu_fw:
                vpu_fw = self.fwp_manager.get_vpu_fw(session.fwp_checksum)
                logging.debug(f'[SessionManager] Starting session ({id}) - vpu entry: {vpu_fw}')
                fw_target_name = 'luxonis_vpu.bin'
                fw_target_path = '/var/lib/firmware/intel/vpux/' + fw_target_name
                try:
                    subprocess.check_call(['cp', str(vpu_fw), fw_target_path])
                    subprocess.check_call(['StartVpu', str(fw_target_name)])
                except:
                    session.state = 'CRASHED'
                    raise PublicError(f'Session with given id: {id} failed booting VPU')

            logging.debug(f'[SessionManager] Starting session ({id}) - fwp entry: {entry}, has vpu: {has_vpu_fw}, vpu entry: {vpu_fw}, root_folder: {root_folder}')
            fd_watchdog_read, fd_watchdog_write = os.pipe()
            session.watchdog.watchdog_fd = fd_watchdog_read
            # Update the custom environment with the current environment and set the WATCHDOG FD correctly
            if type(custom_env) == dict:
                custom_env.update(os.environ)
            else:
                custom_env = os.environ.copy()

            custom_env["DEPTHAI_GATE_WATCHDOG_FD"] = str(fd_watchdog_write)
            # Set the /proc/sys/kernel/core_pattern to the core dump path
            core_pattern_path = Path("/proc/sys/kernel/core_pattern")
            if core_pattern_path.exists():
                with open(core_pattern_path, 'w') as f:
                    f.write("core.%p")
            else:
                logging.error(f"Cannot set the core pattern {core_pattern_path}")

            logging.debug(f"Environment passed to firmware: \n {custom_env}")
            session.proc = subprocess.Popen([entry], cwd=root_folder, close_fds=True,
                                            pass_fds=[fd_watchdog_read, fd_watchdog_write],
                                            env=custom_env,
                                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                            shell=True)
            # Get the parent path from entry
            parent_dir = os.path.dirname(entry)
            session.log_file_path = Path(parent_dir) / "depthai-gate-stdout.log"
            session.core_dump_path = Path(parent_dir) / ("core." + str(session.proc.pid))
            logging.debug(f"Log path: {session.log_file_path}")
            logging.debug(f"Core dump path: {session.core_dump_path}")
            session.output_collector = OutputCollector(session.proc, session.log_file_path)
            session.output_collector.start()
            os.close(fd_watchdog_write)
            session.watchdog.proc = session.proc
            session.state = 'RUNNING'
            session.watchdog.run()

    def stop(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            session = self.session_map[id]
            if session.proc is None:
                raise PublicError(f'Session with given id: {id} was not started')

            session.stop_datetime = datetime.now()
            if self.on_event_module:
                try:
                    self.on_event_module.on_stop()
                except Exception as e:
                    logging.error("Exception thrown on calling the on_stop function {e}")
            session.proc.terminate()
            session.state = 'STOPPING'

    def destroy(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            session = self.session_map[id]

            if self.on_event_module:
                try:
                    self.on_event_module.on_destroy()
                except Exception as e:
                    logging.error("Exception thrown on calling the on_destroy function {e}")

            session.watchdog.stop()
            if session.proc is not None:
                session.proc.kill()
            session.destroy_datetime = datetime.now()
            session.state = 'DESTROYED'

    def refresh_process_state(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            session : Session = self.session_map[id]
            if session.proc is not None:
                retcode = session.proc.poll()
                logging.debug(f'[SessionManager] Session ({id}) process poll: {retcode}')
                if retcode is not None:
                    session.state = 'STOPPED'
                    if retcode != 0:
                        session.state = 'CRASHED'
            else:
                logging.debug(f'[SessionManager] Session ({id}) process not started...')

    def get(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            self.refresh_process_state(id)
            return self.session_map[id]

    def get_all(self):
        with self.mutex:
            sessions = []
            for sessionId in self.session_map:
                sessions.append(self.get(sessionId))
            return sessions

    def delete(self, id):
        with self.mutex:
            if id not in self.session_map:
                raise PublicError(f'Session with given id: {id} does not exist')
            self.destroy(id)
            del self.session_map[id]
