import subprocess
import logging

def is_linked(binary_path, library):
    try:
        # Use strings and grep to check if the library is linked to the binary
        grep_command = ['grep', library]
        strings_command = ['strings', binary_path]

        # Use subprocess to pipe strings output to grep input
        with subprocess.Popen(strings_command, stdout=subprocess.PIPE) as strings_proc:
            with subprocess.Popen(grep_command, stdin=strings_proc.stdout, stdout=subprocess.PIPE, text=True) as grep_proc:
                strings_proc.stdout.close()  # Allow strings_proc to receive a SIGPIPE if grep_proc exits
                output, _ = grep_proc.communicate()  # Get the output from grep
                return bool(output)  # Return True if the output contains the library string, otherwise False

    except subprocess.CalledProcessError as e:
        # Handle the case where the command cannot process the binary
        logging.error(f"Failed to process binary {binary_path}: {e}")
        return False
    except FileNotFoundError:
        # Handle the case where the binary or tools do not exist
        logging.error(f"Tool or binary not found: {binary_path}")
        return False


def has_hailo():
    # root@keembay:~# hailortcli scan
    # Hailo Devices:
    # [-] Device: 0000:01:00.0

    # Run hailortcli scan
    result = subprocess.run(['hailortcli', 'scan'], text=True, capture_output=True)
    # Check if successful
    if result.returncode != 0:
        print(f"Failed to run hailortcli scan: {result.stderr}")
        return False
    if "Device:" in result.stdout:
        return True
    else:
        return False


def hailo_compatible(binary_path):
    # If hailo is not present on the device, always return True
    if not has_hailo():
        logging.info(f"Device doesn't have Hailo - always compatbile")
        return True

    if is_linked(binary_path, 'libhailort.so.4.17.0'):
        logging.info(f'{binary_path} is compatible with Hailo')
        return True
    elif is_linked(binary_path, 'libhailort.so.4.10.0'):
        logging.warning(f'{binary_path} is not compatible with Hailo')
        return False
    elif is_linked(binary_path, 'libhailort.so\n'):
        # Shouldn't get here
        logging.warning(f'{binary_path} might not be compatible with Hailo')
        return False
    else:
        logging.info(f"{binary_path} doesn't link to Hailo")
        return True
