import sys
import time
import os
import select
import subprocess
from threading import Thread, RLock, Condition
import logging

TIME_TO_SLEEP = 8  # Seconds - XLINK_TCP_WATCHDOG_TIMEOUT is relevant here, but let more time on gate to allow a clean shutdown of the process
INITIAL_DELAY = 10  # Seconds - should match DEPTHAI_WATCHDOG_INITIAL_DELAY

class Watchdog:
    running: bool
    counter: int
    last_counter: int
    time_to_sleep: float
    mutex: RLock
    condition: Condition
    watchdog_thread: Thread
    read_thread: Thread
    watchdog_fd: int
    proc: subprocess.Popen
    session_id: int

    def __init__(self, id):
        self.counter = 1
        self.last_counter = 0
        self.time_to_sleep = TIME_TO_SLEEP
        self.running = False
        self.mutex = RLock()
        self.condition = Condition(self.mutex)
        self.watchdog_thread = Thread(target=self.watchdog_function)
        self.read_thread = Thread(target=self.read_function)
        self.watchdog_fd = None
        self.proc = None
        self.session_id = id

    def increment(self):
        with self.mutex:
            self.counter = self.counter + 1 if self.counter < sys.maxsize else 0

    def read_function(self):
        while self.running:
            with self.condition:
                ready, _, _ = select.select((self.watchdog_fd,), (), (), 0.1)
                if ready:
                    bytes_read = os.read(self.watchdog_fd, 20)
                    if bytes_read.decode() == "ALIVE":
                        self.increment()
                self.condition.wait(timeout=0.1)
        if self.watchdog_fd:
            os.close(self.watchdog_fd)

    def watchdog_function(self):
        with self.condition:
            self.condition.wait(timeout=INITIAL_DELAY)  # Grace period for the app to set up

        while self.running:
            with self.condition:
                if self.counter == self.last_counter:
                    retcode = self.proc.poll() if self.proc else None
                    if retcode is None: # Process still running
                        logging.warning(f"Process on session {self.session_id} not bumping the watchdog. Terminating.")
                        self.kill()
                    else:
                        logging.debug(f'[SessionManager] Session ({self.session_id}) process poll: {retcode}')
                    self.running = False
                else:
                    self.last_counter = self.counter

                self.condition.wait(timeout=self.time_to_sleep)

    def kill(self):
        with self.condition:
            self.running = False
            self.condition.notify_all()

        if self.proc:
            self.proc.terminate()
            try:
                self.proc.wait(timeout=10)
            except subprocess.TimeoutExpired:
                self.proc.kill()

    def run(self):
        logging.info("Starting the watchdog")
        if self.proc is None:
            logging.warning("No process given to the watchdog object. Watchdog will be disabled!")
            return
        if self.watchdog_fd is None:
            logging.warning("No file descriptor given to the watchdog object. Watchdog will be disabled! ")
            return
        self.running = True
        self.watchdog_thread.start()
        self.read_thread.start()

    def stop(self):
        with self.condition:
            self.running = False
            self.condition.notify_all()  # Interrupt any sleeping threads

        try:
            self.watchdog_thread.join()
            self.read_thread.join()
        except RuntimeError as e: # Raised if the thread is not started
            logging.error(f"Error stopping watchdog: {e}")
